#
# Copyright 2017 National Renewable Energy Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Created on 03/09/2015
@author: MMPE
Copied from https://github.com/WISDEM/AeroelasticSE/tree/openmdao1/src/AeroelasticSE/old_files on 15 Aug 2016 by Ganesh Vijayakumar
'''
import os
import numpy as np
import struct

class FASTOutputFile():
    """ 
    Read an OpenFAST ouput file (.out, .outb, .elev).

    Main methods
    ------------
    - read, write, toDataFrame
    """

    def __init__(self, filename=None, method='numpy'):
        """
        Load a FAST binary or ascii output file

        Parameters
        ----------
        filename : str
            filename

        Returns
        -------
        data: ndarray
            data values
        info: dict
            info containing:
                - name: filename
                - description: description of dataset
                - attribute_names: list of attribute names
                - attribute_units: list of attribute units
        """

        assert os.path.isfile(filename), "File, %s, does not exists" % filename

        ext = os.path.splitext(filename)[1]
        if ext in ['.out']:
            with open(filename, 'r') as f:
                try:
                    f.readline()
                except UnicodeDecodeError:
                    self.data, self.info = load_binary_output(filename)

            self.data, self.info = load_ascii_output(filename)

        elif ext == '.outb':
            self.data, self.info, self.pack = load_binary_output(filename)

        else:
            raise ValueError(f'File extension, {ext}, not supported')
        
        if method == 'pandas':
            self.toDataFrame()
            


    def toDataFrame(self):
        import pandas as pd
        """
        Returns object into one DataFrame, or a dictionary of DataFrames
        Borrowed from openfast_toolbox
        """
        # --- Example (returning one DataFrame):
        #  return pd.DataFrame(data=np.zeros((10,2)),columns=['Col1','Col2'])
        if self.info['attribute_units'] is not None:
            if len(self.info['attribute_names'])!=len(self.info['attribute_units']):
                cols=self.info['attribute_names']
                print('[WARN] not all columns have units! Skipping units')
            else:
                cols=[n+'_['+u.replace('sec','s')+']' for n,u in zip(self.info['attribute_names'],self.info['attribute_units'])]
        else:
            cols=self.info['attribute_names']
        if isinstance(self.data, pd.DataFrame):
            df= self.data
            df.columns=cols
        else:
            if len(cols)!=self.data.shape[1]:
                raise Warning('Inconstistent number of columns between headers ({}) and data ({}) for file {}'.format(len(cols), self.data.shape[1], self.filename))
            df = pd.DataFrame(data=self.data,columns=cols)

        return df


def load_ascii_output(filename, headerLines = 8, descriptionLine = 4, attributeLine = 6, unitLine = 7, delimiter = None):
    with open(filename) as f:
        info = {}
        info['name'] = os.path.splitext(os.path.basename(filename))[0]
        header = [f.readline() for _ in range(headerLines)]
        info['description'] = header[descriptionLine].strip()
        info['attribute_names'] = header[attributeLine].strip().split(delimiter)
        info['attribute_units'] = [unit[1:-1] for unit in header[unitLine].strip().split(delimiter)]  #removing "()"
        data = np.array([line.split(delimiter) for line in f.readlines()], dtype=float)
        return data, info

def load_binary_output(filename):
    """
    Ported from ReadFASTbinary.m by Mads M Pedersen, DTU Wind
    Info about ReadFASTbinary.m:
    Author: Bonnie Jonkman, National Renewable Energy Laboratory
    (c) 2012, National Renewable Energy Laboratory
    Edited for FAST v7.02.00b-bjj  22-Oct-2012
    """

    def fread(fid, n, type):
        fmt, nbytes = {'uint8': ('B', 1), 'int16':('h', 2), 'int32':('i', 4), 'float32':('f', 4), 'float64':('d', 8)}[type]
        return struct.unpack(fmt * n, fid.read(nbytes * n))
    
    FileFmtID_WithTime = 1    # File identifiers used in FAST
    FileFmtID_WithoutTime = 2
    FileFmtID_NoCompressWithoutTime = 3
    FileFmtID_ChanLen_In = 4
    
    with open(filename, 'rb') as fid:
        FileID = fread(fid, 1, 'int16')[0]       # FAST output file format, INT(2)
        
        if FileID == FileFmtID_ChanLen_In:
            LenName  = fread(fid, 1, 'int16')[0] # Number of characters in channel names and units
        else:
            LenName = 10                         # default number of characters per channel name

        
        NumOutChans = fread(fid, 1, 'int32')[0]  # The number of output channels, INT(4)
        NT = fread(fid, 1, 'int32')[0]           # The number of time steps, INT(4)
        
        if FileID == FileFmtID_WithTime:
            TimeScl = fread(fid, 1, 'float64')   # The time slopes for scaling, REAL(8)
            TimeOff = fread(fid, 1, 'float64')   # The time offsets for scaling, REAL(8)
        else:
            TimeOut1 = fread(fid, 1, 'float64')  # The first time in the time series, REAL(8)
            TimeIncr = fread(fid, 1, 'float64')  # The time increment, REAL(8)

        if FileID != FileFmtID_NoCompressWithoutTime:
            ColScl = fread(fid, NumOutChans, 'float32')  # The channel slopes for scaling, REAL(4)
            ColOff = fread(fid, NumOutChans, 'float32')  # The channel offsets for scaling, REAL(4)

        LenDesc = fread(fid, 1, 'int32')[0]          # The number of characters in the description string, INT(4)
        DescStrASCII = fread(fid, LenDesc, 'uint8')  # DescStr converted to ASCII
        DescStr = "".join(map(chr, DescStrASCII)).strip()

        ChanName = []                                     # initialize the ChanName cell array
        for iChan in range(NumOutChans + 1):
            ChanNameASCII = fread(fid, LenName, 'uint8')  # ChanName converted to numeric ASCII
            ChanName.append("".join(map(chr, ChanNameASCII)).strip())

        ChanUnit = []                                     # initialize the ChanUnit cell array
        for iChan in range(NumOutChans + 1):
            ChanUnitASCII = fread(fid, LenName, 'uint8')  # ChanUnit converted to numeric ASCII
            ChanUnit.append("".join(map(chr, ChanUnitASCII)).strip()[1:-1])

        # get the channel time series
        nPts = NT * NumOutChans                   # number of data points in the file
        if FileID == FileFmtID_WithTime:
            PackedTime = fread(fid, NT, 'int32')  # read the time data
            cnt = len(PackedTime)
            if cnt < NT:
                raise Exception('Could not read entire %s file: read %d of %d time values' % (filename, cnt, NT))
        
        if FileID == FileFmtID_NoCompressWithoutTime:
            PackedData = fread(fid, nPts, 'float64')    # read the channel data
        else:
            PackedData = fread(fid, nPts, 'int16')    # read the channel data
            
        cnt = len(PackedData)
        if cnt < nPts:
            raise Exception('Could not read entire %s file: read %d of %d values' % (filename, cnt, nPts))

    if FileID == FileFmtID_NoCompressWithoutTime:
        pack = np.array(PackedData).reshape(NT, NumOutChans)
        data = pack
    else:
        # Scale the packed binary to real data
        pack = np.array(PackedData).reshape(NT, NumOutChans)
        data = (pack - ColOff) / ColScl

    if FileID == FileFmtID_WithTime:
        time = (np.array(PackedTime) - TimeOff) / TimeScl;
    else:
        time = TimeOut1 + TimeIncr * np.arange(NT)

    data = np.concatenate([time.reshape(NT, 1), data], 1)
    pack = np.concatenate([time.reshape(NT, 1), pack], 1)

    info = {'name': os.path.splitext(os.path.basename(filename))[0],
            'description': DescStr,
            'attribute_names': ChanName,
            'attribute_units': ChanUnit}
    return data, info, pack



if __name__=="__main__":

    from openfast_io.FileTools import check_rtest_cloned

    parent_dir = os.path.dirname( os.path.dirname( os.path.dirname( os.path.realpath(__file__) ) ) ) + os.sep

    of_outputfile = os.path.join(parent_dir, 'reg_tests', 'r-test', 'glue-codes', 
                                 'openfast', '5MW_Land_BD_DLL_WTurb', '5MW_Land_BD_DLL_WTurb.outb')

    check_rtest_cloned(of_outputfile)

    d,i,p = load_binary_output(of_outputfile)

    print(tuple(i['attribute_names']))
    print(type(d))
    print(i)
    print(len(i['attribute_names']))
    print(np.shape(d))



def default_StC_vt():
    StC = {}

    StC['Echo'] = True

    #---------------------- StC DEGREES OF FREEDOM ----------------------------------
    StC['StC_DOF_MODE'] = 0
    StC['StC_X_DOF']    = False
    StC['StC_Y_DOF']    = False
    StC['StC_Z_DOF']    = False

    # ---------------------- StC LOCATION ------------------------------------------- [relative to the reference origin of component attached to]
    StC['StC_P_X']  = 0     #      - At rest X position of StC (m)
    StC['StC_P_Y']  = 0     #      - At rest Y position of StC (m)
    StC['StC_P_Z']  = 0     #      - At rest Z position of StC (m)
    # ---------------------- StC INITIAL CONDITIONS --------------------------------- [used only when StC_DOF_MODE=1 or 2]
    StC['StC_X_DSP'] = 0    #    - StC X initial displacement (m) [relative to at rest position]
    StC['StC_Y_DSP'] = 0    #    - StC Y initial displacement (m) [relative to at rest position]
    StC['StC_Z_DSP'] = 0    #    - StC Z initial displacement (m) [relative to at rest position; used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_Z_PreLd'] = "none"    #    - StC Z initial displacement (m) [relative to at rest position; used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    # ---------------------- StC CONFIGURATION -------------------------------------- [used only when StC_DOF_MODE=1 or 2]
    StC['StC_X_PSP'] = 0    #    - Positive stop position (maximum X mass displacement) (m)
    StC['StC_X_NSP'] = 0    #    - Negative stop position (minimum X mass displacement) (m)
    StC['StC_Y_PSP'] = 0    #    - Positive stop position (maximum Y mass displacement) (m)
    StC['StC_Y_NSP'] = 0    #    - Negative stop position (minimum Y mass displacement) (m)
    StC['StC_Z_PSP'] = 0    #    - Positive stop position (maximum Z mass displacement) (m) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_Z_NSP'] = 0    #    - Negative stop position (minimum Z mass displacement) (m) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    #---------------------- StC MASS, STIFFNESS, & DAMPING ------------------------- [used only when StC_DOF_MODE=1 or 2]
    StC['StC_X_M']      = 0     #  - StC X mass (kg) [must equal StC_Y_M for StC_DOF_MODE = 2]
    StC['StC_Y_M']      = 0     #  - StC Y mass (kg) [must equal StC_X_M for StC_DOF_MODE = 2]
    StC['StC_Z_M']      = 0     #  - StC Z mass (kg) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_XY_M']     = 0     #  - StC Z mass (kg) [used only when StC_DOF_MODE=2]
    StC['StC_X_K']      = 0     #  - StC X stiffness (N/m)
    StC['StC_Y_K']      = 0     #  - StC Y stiffness (N/m)
    StC['StC_Z_K']      = 0     #  - StC Z stiffness (N/m) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_X_C']      = 0     #  - StC X damping (N/(m/s))
    StC['StC_Y_C']      = 0     #  - StC Y damping (N/(m/s))
    StC['StC_Z_C']      = 0     #  - StC Z damping (N/(m/s)) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_X_KS']     = 0     #  - Stop spring X stiffness (N/m)
    StC['StC_Y_KS']     = 0     #  - Stop spring Y stiffness (N/m)
    StC['StC_Z_KS']     = 0     #  - Stop spring Z stiffness (N/m) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_X_CS']     = 0     #  - Stop spring X damping (N/(m/s))
    StC['StC_Y_CS']     = 0     #  - Stop spring Y damping (N/(m/s))
    StC['StC_Z_CS']     = 0     #  - Stop spring Z damping (N/(m/s)) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    #---------------------- StC USER-DEFINED SPRING FORCES ------------------------- [used only when StC_DOF_MODE=1 or 2]
    StC['Use_F_TBL']    = False   # - Use spring force from user-defined table (flag)
    StC['NKInpSt']      = 2       # - Number of spring force input stations
    #---------------------- StC SPRING FORCES TABLE -------------------------------- [used only when StC_DOF_MODE=1 or 2]
    StC['SpringForceTable'] = {}
    StC['SpringForceTable']['X']        =  [-6,6]
    StC['SpringForceTable']['F_X']      =  [-4.8e6,4.8e6]
    StC['SpringForceTable']['Y']        =  [-6,6]
    StC['SpringForceTable']['F_Y']      =  [-4.8e6,4.8e6]
    StC['SpringForceTable']['Z']        =  [-6,6]
    StC['SpringForceTable']['F_Z']      =  [-4.8e6,4.8e6]
    # ---------------------- StructCtrl CONTROL -------------------------------------------- [used only when StC_DOF_MODE=1 or 2]
    StC['StC_CMODE']        = 0       # - Control mode (switch) {0:none; 1: Semi-Active Control Mode; 2: Active Control Mode}
    StC['StC_SA_MODE']      = 1       # - Semi-Active control mode {1: velocity-based ground hook control; 2: Inverse velocity-based ground hook control; 3: displacement-based ground hook control 4: Phase difference Algorithm with Friction Force 5: Phase difference Algorithm with Damping Force} (-)
    StC['StC_CChan']        = 0       # - Control mode (switch) {0:none; 1: Semi-Active Control Mode; 2: Active Control Mode}
    StC['StC_X_C_HIGH']     = 0       # - StC X high damping for ground hook control
    StC['StC_X_C_LOW']      = 0       # - StC X low damping for ground hook control
    StC['StC_Y_C_HIGH']     = 0       # - StC Y high damping for ground hook control
    StC['StC_Y_C_LOW']      = 0       # - StC Y low damping for ground hook control
    StC['StC_Z_C_HIGH']     = 0       # - StC Z high damping for ground hook control [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_Z_C_LOW']      = 0       # - StC Z low damping for ground hook control  [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    StC['StC_X_C_BRAKE']    = 0       # - StC X high damping for braking the StC (Don't use it now. should be zero)
    StC['StC_Y_C_BRAKE']    = 0       # - StC Y high damping for braking the StC (Don't use it now. should be zero)
    StC['StC_Z_C_BRAKE']    = 0       # - StC Z high damping for braking the StC (Don't use it now. should be zero) [used only when StC_DOF_MODE=1 and StC_Z_DOF=TRUE]
    # ---------------------- TLCD --------------------------------------------------- [used only when StC_DOF_MODE=3]
    StC['L_X']              = 0     # - X TLCD total length (m)
    StC['B_X']              = 0     # - X TLCD horizontal length (m)
    StC['area_X']           = 0     # - X TLCD cross-sectional area of vertical column (m^2)
    StC['area_ratio_X']     = 0     # - X TLCD cross-sectional area ratio (vertical column area divided by horizontal column area) (-)
    StC['headLossCoeff_X']  = 0     # - X TLCD head loss coeff (-)
    StC['rho_X']            = 0     # - X TLCD liquid density (kg/m^3)
    StC['L_Y']              = 0     # - Y TLCD total length (m)
    StC['B_Y']              = 0     # - Y TLCD horizontal length (m)
    StC['area_Y']           = 0     # - Y TLCD cross-sectional area of vertical column (m^2)
    StC['area_ratio_Y']     = 0     # - Y TLCD cross-sectional area ratio (vertical column area divided by horizontal column area) (-)
    StC['headLossCoeff_Y']  = 0     # - Y TLCD head loss coeff (-)
    StC['rho_Y']            = 0     # - Y TLCD liquid density (kg/m^3)
    # ---------------------- PRESCRIBED TIME SERIES --------------------------------- [used only when StC_DOF_MODE=4]
    StC['PrescribedForcesCoord']    = 0                 # - Prescribed forces are in global or local coordinates (switch) {1: global; 2: local}
    StC['PrescribedForcesFile']     = "unused"          #  - Time series force and moment (7 columns of time, FX, FY, FZ, MX, MY, MZ)

    return StC


if __name__=="__main__":
    
    a = default_StC_vt()
    print('here')




import os
import xlrd

""" 
Script to generate the Output channel dictionaries.  Included so they can be easily updated
for future FAST versions and additional FAST modules.

The script did not work for me with a xlsx worksheet, but it did with a legacy xls file

by Evan Gaertner, maintained by Pietro Bortolotti @ NREL on June 2nd, 2021
"""


def GetOutlistParameters(fname_vars_out, xl_files_in, sheet_list, write_mode, final_dict):

    f = open(fname_vars_out, write_mode)
    f.write('""" Generated from FAST OutListParameters.xlsx files with AeroelasticSE/src/AeroelasticSE/Util/create_output_vars.py """\n')

    for fname, sheet_name in zip(xl_files_in, sheet_list):
        book = xlrd.open_workbook(fname)
        sheet_idx = book.sheet_names().index(sheet_name)
        sheet = book.sheet_by_index(sheet_idx)

        print(sheet_name, '\t:\t', fname)

        # get data from sheet
        outlist = {}
        headers = {}
        var_names = []
        n = sheet.nrows
        n_dup = 0

        for idx in range(1, sheet.nrows):
            row = sheet.row_values(idx)
            if row[1]!='' and row[5]!='':
                outvar = str(row[1])
                var_names.append(str(row[1]))
                outlist[outvar] = {}
                outlist[outvar]['desc'] = ''.join([i if ord(i) < 128 else ' ' for i in row[3]])
                outlist[outvar]['conv'] = ''.join([i if ord(i) < 128 else ' ' for i in row[4]])
                outlist[outvar]['unit'] = ''.join([i if ord(i) < 128 else ' ' for i in row[5]])
                outlist[outvar]['idx'] = idx

                for var in row[2].split(','):
                    if var.strip() != '':
                        outlist[str(var.strip())] = outlist[outvar]
                        var_names.append(str(var.strip()))
                        n += 1
                        n_dup +=1

            else:
                headers[idx+n_dup] = str(row[0])
        
        # write out variables + metadata
        f.write('\n\n""" ' + sheet_name + ' """\n')
        f.write(sheet_name + ' = {}\n')

        var_count = 0
        for idx in range(1, n):
            if idx in headers.keys():
                f.write('\n# ' + headers[idx] + '\n')
            else:
                var = var_names[var_count]
                f.write(("%s['%s']"%(sheet_name, var)).ljust(30) + "= False     # %s; %s; %s\n"%(outlist[var]['unit'], outlist[var]['desc'], outlist[var]['conv']))
                var_count += 1

    # combine in one dict
    f.write('\n\n""" Final Output Dictionary """\n')
    f.write('%s = {}\n'%final_dict)
    for var in sheet_list:
        f.write(("%s['%s']"%(final_dict, var)).ljust(30) + "= %s\n"%var)


    f.close()


if __name__=="__main__":

    try:
        # File Location
        root_dir = os.path.dirname(os.path.dirname( os.path.dirname( ( os.path.realpath(__file__) ) ) )) + os.sep
        outlist_fast_lib = os.path.join(root_dir, 'docs', 'OtherSupporting' , 'OutListParameters.xls')
        # Sheets to grab
        sheet_list = [
                        'AeroDyn', 
                        'BeamDyn', 
                        'ElastoDyn', 
                        'InflowWind', 
                        'ServoDyn', 
                        'HydroDyn', 
                        'Morison', 
                        'SeaState', 
                        'SubDyn',
                        'AeroDyn_Nodes',
                        'BeamDyn_Nodes',
                        'ElastoDyn_Nodes',
                        'AeroDisk',
                        'SimpleElastoDyn',
                        ] 
        xl_files = [outlist_fast_lib]*len(sheet_list)
        # Output naming
        fname_vars_out = 'FAST_vars_out.py'
        # Run
        GetOutlistParameters(fname_vars_out, xl_files, sheet_list, 'w', 'FstOutput')
    except:
        raise Exception("Try saving the OpenFAST xlsx file into a legacy xls file")


""" 
Taken from python_toolbox https://github.com/openfast/python_toolbox
"""
import os
import numpy as np
import re
class BrokenFormatError(Exception): pass

class FASTLinearizationFile(dict):
    """ 
    Read/write an OpenFAST linearization file. The object behaves like a dictionary.

    Main keys
    ---------
    - 'x', 'xdot' 'u', 'y', 'A', 'B', 'C', 'D'

    Main methods
    ------------
    - read, write, toDataFrame, keys, xdescr, ydescr, udescr

    Examples
    --------

        f = FASTLinearizationFile('5MW.1.lin')
        print(f.keys())
        print(f['u'])     # input operating point
        print(f.udescr()) # description of inputs

        # use a dataframe with "named" columns and rows
        df = f.toDataFrame()
        print(df['A'].columns)
        print(df['A'])

    """
    @staticmethod
    def defaultExtensions():
        return ['.lin']

    @staticmethod
    def formatName():
        return 'FAST linearization output'

    def __init__(self, filename=None, **kwargs):
        """ Class constructor. If a `filename` is given, the file is read. """
        self.filename = filename
        if filename:
            self.read(**kwargs)

    def read(self, filename=None, **kwargs):
        """ Reads the file self.filename, or `filename` if provided """
        
        # --- Standard tests and exceptions (generic code)
        if filename:
            self.filename = filename
        if not self.filename:
            raise Exception('No filename provided')
        if not os.path.isfile(self.filename):
            raise OSError(2,'File not found:',self.filename)
        if os.stat(self.filename).st_size == 0:
            raise EmptyFileError('File is empty:',self.filename)
        # --- Calling (children) function to read
        self._read(**kwargs)

    def _read(self, *args, **kwargs):
        self['header']=[]

        def extractVal(lines, key):
            for l in lines:
                if l.find(key)>=0:
                    return l.split(key)[1].split()[0]
            return None

        def readToMarker(fid, marker, nMax):
            lines=[]
            for i, line in enumerate(fid):
                if i>nMax:
                    raise BrokenFormatError('`{}` not found in file'.format(marker))
                if line.find(marker)>=0:
                    break
                lines.append(line.strip())
            return lines, line
        
        def readOP(fid, n, name=''):
            OP=[]
            Var = {'RotatingFrame': [], 'DerivativeOrder': [], 'Description': []}
            colNames=fid.readline().strip()
            fid.readline().strip()
            bHasDeriv= colNames.find('Derivative Order')>=0
            for i, line in enumerate(fid):
                sp = line.strip().split()
                if sp[1].find(',') >= 0:
                    #  Most likely this OP has three values (e.g. orientation angles)
                    # For now we discard the two other values
                    OP.append(np.float32(sp[1][:-1]))
                    iRot=4
                else:
                    OP.append(np.float32(sp[1]))
                    iRot=2
                Var['RotatingFrame'].append(sp[iRot])
                if bHasDeriv:
                    Var['DerivativeOrder'].append(int(sp[iRot+1]))
                    Var['Description'].append(' '.join(sp[iRot+2:]).strip())
                else:
                    Var['DerivativeOrder'].append(-1)
                    Var['Description'].append(' '.join(sp[iRot+1:]).strip())
                if i>=n-1:
                    break
            OP=np.asarray(OP)
            return OP, Var

        def readMat(fid, n, m, name=''):

            # Read rows from file, raise exception on failure
            try:
                vals = np.genfromtxt(fid, dtype=np.float64, max_rows=n)
            except:
                raise Exception('Failed to convert into an array of float the matrix `{}`\n\tin linfile: {}'.format(name, self.filename))
            
            # Raise exception if actual matrix shape does not match expected shape
            if vals.shape[0]!=n or vals.shape[1]!=m:
                shape1 = vals.shape
                shape2 = (n,m)
                raise Exception('Shape of matrix `{}` has wrong dimension ({} instead of {})\n\tin linfile: {}'.format(name, shape1, shape2, name, self.filename))

            # Raise exceptions if any elements are NaN or infinity
            if np.any(np.isnan(vals.ravel())):
                raise Exception('Some ill-formated/infinite values (e.g. `*******`) were found in the matrix `{}`\n\tin linflile: {}'.format(name, self.filename))
            if np.any(np.isinf(vals.ravel())):
                raise Exception('Some NaN values were found in the matrix `{}`\n\tin linfile: `{}`.'.format(name, self.filename))
            return vals


        # Reading 
        with open(self.filename, 'r', errors="surrogateescape") as f:
            # --- Reader header
            self['header'], lastLine=readToMarker(f, 'Jacobians included', 30)
            self['header'].append(lastLine)
            nx  = int(extractVal(self['header'],'Number of continuous states:'))
            nxd = int(extractVal(self['header'],'Number of discrete states:'  ))
            nz  = int(extractVal(self['header'],'Number of constraint states:'))
            nu  = int(extractVal(self['header'],'Number of inputs:'           ))
            ny  = int(extractVal(self['header'],'Number of outputs:'          ))
            bJac = extractVal(self['header'],'Jacobians included in this file?')
            try:
                self['Azimuth'] = np.float32(extractVal(self['header'],'Azimuth:'))
            except:
                self['Azimuth'] = None
            try:
                self['RotSpeed'] = np.float32(extractVal(self['header'],'Rotor Speed:')) # rad/s
            except:
                self['RotSpeed'] = None
            try:
                self['WindSpeed'] = np.float32(extractVal(self['header'],'Wind Speed:'))
            except:
                self['WindSpeed'] = None

            for i, line in enumerate(f):
                line = line.strip()
                if line.find('Order of continuous states:')>=0:
                    self['x'], self['x_info'] = readOP(f, nx, 'x')
                elif line.find('Order of continuous state derivatives:')>=0:
                    self['xdot'], self['xdot_info'] = readOP(f, nx, 'xdot')
                elif line.find('Order of inputs')>=0:
                    self['u'], self['u_info'] = readOP(f, nu, 'u')
                elif line.find('Order of outputs')>=0:
                    self['y'], self['y_info'] = readOP(f, ny, 'y')
                elif line.find('A:')>=0:
                    self['A'] = readMat(f, nx, nx, 'A')
                elif line.find('B:')>=0:
                    self['B'] = readMat(f, nx, nu, 'B')
                elif line.find('C:')>=0:
                    self['C'] = readMat(f, ny, nx, 'C')
                elif line.find('D:')>=0:
                    self['D'] = readMat(f, ny, nu, 'D')
                elif line.find('dUdu:')>=0:
                    self['dUdu'] = readMat(f, nu, nu,'dUdu')
                elif line.find('dUdy:')>=0:
                    self['dUdy'] = readMat(f, nu, ny,'dUdy')
                elif line.find('StateRotation:')>=0:
                    pass
                    # TODO
                    #StateRotation:
                elif line.find('ED M:')>=0:
                    self['EDDOF'] = line[5:].split()
                    self['M']     = readMat(f, 24, 24,'M')

    def toString(self):
        s=''
        return s

    def _write(self):
        with open(self.filename,'w') as f:
            f.write(self.toString())

    def short_descr(self,slist):
        def shortname(s):
            s=s.strip()
            s = s.replace('(m/s)'   , '_[m/s]'  );
            s = s.replace('(kW)'    , '_[kW]'   );
            s = s.replace('(deg)'   , '_[deg]'  );
            s = s.replace('(N)'     , '_[N]'    );
            s = s.replace('(kN-m)'  , '_[kNm]' );
            s = s.replace('(N-m)'  , '_[Nm]' );
            s = s.replace('(kN)'  , '_[kN]' );
            s = s.replace('(rpm)'   , '_[rpm]'  );
            s = s.replace('(rad)'   , '_[rad]'  );
            s = s.replace('(rad/s)' , '_[rad/s]'  );
            s = s.replace('(rad/s^2)', '_[rad/s^2]'  );
            s = s.replace('(m/s^2)' , '_[m/s^2]');
            s = s.replace('(deg/s^2)','_[deg/s^2]');
            s = s.replace('(m)'     , '_[m]'    );
            s = s.replace(', m/s/s','_[m/s^2]');
            s = s.replace(', m/s^2','_[m/s^2]');
            s = s.replace(', m/s','_[m/s]');
            s = s.replace(', m','_[m]');
            s = s.replace(', rad/s/s','_[rad/s^2]');
            s = s.replace(', rad/s^2','_[rad/s^2]');
            s = s.replace(', rad/s','_[rad/s]');
            s = s.replace(', rad','_[rad]');
            s = s.replace(', -','_[-]');
            s = s.replace(', Nm/m','_[Nm/m]');
            s = s.replace(', Nm','_[Nm]');
            s = s.replace(', N/m','_[N/m]');
            s = s.replace(', N','_[N]');
            s = s.replace('(1)','1')
            s = s.replace('(2)','2')
            s = s.replace('(3)','3')
            s= re.sub(r'\([^)]*\)','', s) # remove parenthesis
            s = s.replace('ED ','');
            s = s.replace('BD_','BD_B');
            s = s.replace('IfW ','');
            s = s.replace('Extended input: ','')
            s = s.replace('1st tower ','qt1');
            s = s.replace('2nd tower ','qt2');
            nd = s.count('First time derivative of ')
            if nd>=0:
                s = s.replace('First time derivative of '     ,'');
                if nd==1:
                    s = 'd_'+s.strip()
                elif nd==2:
                    s = 'dd_'+s.strip()
            s = s.replace('Variable speed generator DOF ','psi_rot'); # NOTE: internally in FAST this is the azimuth of the rotor
            s = s.replace('fore-aft bending mode DOF '    ,'FA'     );
            s = s.replace('side-to-side bending mode DOF','SS'     );
            s = s.replace('bending-mode DOF of blade '    ,''     );
            s = s.replace(' rotational-flexibility DOF, rad','-ROT'   );
            s = s.replace('rotational displacement in ','rot'   );
            s = s.replace('Drivetrain','DT'   );
            s = s.replace('translational displacement in ','trans'   );
            s = s.replace('finite element node ','N'   );
            s = s.replace('-component position of node ','posN')
            s = s.replace('-component inflow on tower node','TwrN')
            s = s.replace('-component inflow on blade 1, node','Bld1N')
            s = s.replace('-component inflow on blade 2, node','Bld2N')
            s = s.replace('-component inflow on blade 3, node','Bld3N')
            s = s.replace('-component inflow velocity at node','N')
            s = s.replace('X translation displacement, node','TxN')
            s = s.replace('Y translation displacement, node','TyN')
            s = s.replace('Z translation displacement, node','TzN')
            s = s.replace('X translation velocity, node','TVxN')
            s = s.replace('Y translation velocity, node','TVyN')
            s = s.replace('Z translation velocity, node','TVzN')
            s = s.replace('X translation acceleration, node','TAxN')
            s = s.replace('Y translation acceleration, node','TAyN')
            s = s.replace('Z translation acceleration, node','TAzN')
            s = s.replace('X orientation angle, node'  ,'RxN')
            s = s.replace('Y orientation angle, node'  ,'RyN')
            s = s.replace('Z orientation angle, node'  ,'RzN')
            s = s.replace('X rotation velocity, node'  ,'RVxN')
            s = s.replace('Y rotation velocity, node'  ,'RVyN')
            s = s.replace('Z rotation velocity, node'  ,'RVzN')
            s = s.replace('X rotation acceleration, node'  ,'RAxN')
            s = s.replace('Y rotation acceleration, node'  ,'RAyN')
            s = s.replace('Z rotation acceleration, node'  ,'RAzN')
            s = s.replace('X force, node','FxN')
            s = s.replace('Y force, node','FyN')
            s = s.replace('Z force, node','FzN')
            s = s.replace('X moment, node','MxN')
            s = s.replace('Y moment, node','MyN')
            s = s.replace('Z moment, node','MzN')
            s = s.replace('FX', 'Fx')
            s = s.replace('FY', 'Fy')
            s = s.replace('FZ', 'Fz')
            s = s.replace('MX', 'Mx')
            s = s.replace('MY', 'My')
            s = s.replace('MZ', 'Mz')
            s = s.replace('FKX', 'FKx')
            s = s.replace('FKY', 'FKy')
            s = s.replace('FKZ', 'FKz')
            s = s.replace('MKX', 'MKx')
            s = s.replace('MKY', 'MKy')
            s = s.replace('MKZ', 'MKz')
            s = s.replace('Nodes motion','')
            s = s.replace('cosine','cos'   );
            s = s.replace('sine','sin'   );
            s = s.replace('collective','coll.');
            s = s.replace('Blade','Bld');
            s = s.replace('rotZ','TORS-R');
            s = s.replace('transX','FLAP-D');
            s = s.replace('transY','EDGE-D');
            s = s.replace('rotX','EDGE-R');
            s = s.replace('rotY','FLAP-R');
            s = s.replace('flapwise','FLAP');
            s = s.replace('edgewise','EDGE');
            s = s.replace('horizontal surge translation DOF','Surge');
            s = s.replace('horizontal sway translation DOF','Sway');
            s = s.replace('vertical heave translation DOF','Heave');
            s = s.replace('roll tilt rotation DOF','Roll');
            s = s.replace('pitch tilt rotation DOF','Pitch');
            s = s.replace('yaw rotation DOF','Yaw');
            s = s.replace('vertical power-law shear exponent','alpha')
            s = s.replace('horizontal wind speed ','WS')
            s = s.replace('propagation direction','WD')
            s = s.replace(' pitch command','pitch')
            s = s.replace('HSS_','HSS')
            s = s.replace('Bld','B')
            s = s.replace('tower','Twr')
            s = s.replace('Tower','Twr')
            s = s.replace('Nacelle','Nac')
            s = s.replace('Platform','Ptfm')
            s = s.replace('SrvD','SvD')
            s = s.replace('Generator torque','Qgen')
            s = s.replace('coll. blade-pitch command','PitchColl')
            s = s.replace('wave elevation at platform ref point','WaveElevRefPoint')
            s = s.replace('1)','1');
            s = s.replace('2)','2');
            s = s.replace('3)','3');
            s = s.replace(',','');
            s = s.replace(' ','');
            s=s.strip()
            return s
        return [shortname(s) for s in slist]

    def xdescr(self):
        if 'x_info' in self.keys():
            return self.short_descr(self['x_info']['Description'])
        else:
            return []

    def xdotdescr(self):
        if 'xdot_info' in self.keys():
            return self.short_descr(self['xdot_info']['Description'])
        else:
            return []

    def ydescr(self):
        if 'y_info' in self.keys():
            return self.short_descr(self['y_info']['Description'])
        else:
            return []
    def udescr(self):
        if 'u_info' in self.keys():
            return self.short_descr(self['u_info']['Description'])
        else:
            return []

    def toDataFrame(self):
        import pandas as pd
        dfs={}

        xdescr_short    = self.xdescr()
        xdotdescr_short = self.xdotdescr()
        ydescr_short    = self.ydescr()
        udescr_short    = self.udescr()

        if 'A' in self.keys():
            dfs['A'] = pd.DataFrame(data = self['A'], index=xdescr_short, columns=xdescr_short)
        if 'B' in self.keys():
            dfs['B'] = pd.DataFrame(data = self['B'], index=xdescr_short, columns=udescr_short)
        if 'C' in self.keys():
            dfs['C'] = pd.DataFrame(data = self['C'], index=ydescr_short, columns=xdescr_short)
        if 'D' in self.keys():
            dfs['D'] = pd.DataFrame(data = self['D'], index=ydescr_short, columns=udescr_short)
        if 'x' in self.keys():
            dfs['x'] = pd.DataFrame(data = np.asarray(self['x']).reshape((1,-1)), columns=xdescr_short)
        if 'xdot' in self.keys():
            dfs['xdot'] = pd.DataFrame(data = np.asarray(self['xdot']).reshape((1,-1)), columns=xdotdescr_short)
        if 'u' in self.keys():
            dfs['u'] = pd.DataFrame(data = np.asarray(self['u']).reshape((1,-1)), columns=udescr_short)
        if 'y' in self.keys():
            dfs['y'] = pd.DataFrame(data = np.asarray(self['y']).reshape((1,-1)), columns=ydescr_short)
        if 'M' in self.keys():
            dfs['M'] = pd.DataFrame(data = self['M'], index=self['EDDOF'], columns=self['EDDOF'])
        if 'dUdu' in self.keys():
            dfs['dUdu'] = pd.DataFrame(data = self['dUdu'], index=udescr_short, columns=udescr_short)
        if 'dUdy' in self.keys():
            dfs['dUdy'] = pd.DataFrame(data = self['dUdy'], index=udescr_short, columns=ydescr_short)

        return dfs


